from .resp import Resp
from .multiresp import MultiResp
from .due import due, Doi


def clean_kwargs(func):
    kws = ('stage_2', 'hyp_a1', 'hyp_a2', 'hyp_b', 'restraint', 'basis', 'method', 'ihfree',
           'use_radii', 'solvent')

    def wrapper(*args, **kwargs):
        kwargs = {k: v for k, v in kwargs.items() if k not in kws}
        return func(*args, **kwargs)
    return wrapper


def resp_config(stage_2=True, hyp_a1=0.0005, hyp_a2=0.001,
                hyp_b=0.1, restraint=True, use_radii='msk',
                basis='6-31g*', method='scf', ihfree=True,
                solvent=None):
    """Make analogous classes to R.E.D. options."""
    def wrapper(cls):
        @clean_kwargs
        def run(self, opt=False, save_opt_geometry=False,
                chrconstr=[], chrequiv=[],
                psi4_options={}, n_orient=0, orient=[],
                n_rotate=0, rotate=[], n_translate=0, translate=[],
                equal_methyls=False, **kwargs):
            """
            Perform a Resp fit.

            Parameters
            ----------
            opt: bool (optional)
                Whether to optimise the geometry of each conformer
            save_opt_geometry: bool (optional)
                if ``True``, writes optimised geometries to an XYZ file
            chrconstr: list or dict (optional)
                Intramolecular charge constraints in the form of
                {charge: atom_number_list} or [(charge, atom_number_list)].
                The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
                mean that atoms 1 and 2 together have a charge of 0.
            chrequiv: list (optional)
                lists of atoms with equivalent charges.
                e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
                charges, and atoms 3, 4, and 5 have equal charges.
            psi4_options: dict (optional)
                additional Psi4 options
            n_orient: int (optional)
                If this is greater than 0 and ``orient`` is not given,
                ``n_orient`` orientations are automatically generated for the
                molcule. Heavy atoms are prioritised.
            orient: list of tuples of ints (optional)
                List of reorientations. Corresponds to REMARK REORIENT in R.E.D.
                e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
                around the first, fifth and ninth atom; and the second in reverse
                order.
            n_rotate: int (optional)
                If this is greater than 0 and ``rotate`` is not given,
                ``n_rotate`` rotations are automatically generated for the
                molecule. Heavy atoms are prioritised.
            rotate: list of tuples of ints (optional)
                List of rotations. Corresponds to REMARK ROTATE in R.E.D.
                e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
                around the first, fifth and ninth atom; and the second in reverse
                order.
            n_translate: int (optional)
                If this is greater than 0 and ``translate`` is not given,
                ``n_translate`` translations are randomly generated for the
                molcule in the domain [0, 1).
            translate: list of tuples of floats (optional)
                List of translations. Corresponds to REMARK TRANSLATE in R.E.D.
                e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the
                x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
            equal_methyls: bool (optional)
                if ``True``, all carbons in methyl groups are constrained to be
                equivalent; all carbons in methylenes are equivalent; all hydrogens
                in methyls are equivalent; and all hydrogens in methylenes are
                equivalent. Ignored if ``chrequiv`` constraints are provided.
            **kwargs:
                arguments passed to Resp.fit

            Returns
            -------
            charges: ndarray
            """
            return super(cls, self).run(opt=opt, save_opt_geometry=save_opt_geometry,
                                        chrconstr=chrconstr, chrequiv=chrequiv,
                                        psi4_options=psi4_options,
                                        n_orient=n_orient, orient=orient,
                                        n_rotate=n_rotate, rotate=rotate,
                                        n_translate=n_translate, translate=translate,
                                        equal_methyls=equal_methyls, stage_2=stage_2,
                                        hyp_a1=hyp_a1, hyp_a2=hyp_a2, hyp_b=hyp_b,
                                        restraint=restraint, basis=basis,
                                        method=method, ihfree=ihfree,
                                        solvent=solvent, **kwargs)

        @clean_kwargs
        def optimize_geometry(self, psi4_options={}, save_opt_geometry=False, **kwargs):
            """
            Optimise geometry for all conformers.

            Parameters
            ----------
            psi4_options: dict (optional)
                additional Psi4 options
            save_opt_geometry: bool (optional)
                if ``True``, saves optimised geometries to XYZ file
            """
            return super(cls, self).optimize_geometry(method=method, basis=basis,
                                                      psi4_options=psi4_options, save_opt_geometry=save_opt_geometry)

        @clean_kwargs
        def fit(self, hyp_a=0.0005, tol=1e-6, maxiter=50, **kwargs):
            """
            Perform the Resp fits.

            Parameters
            ----------
            tol: float (optional)
                threshold for convergence
            maxiter: int (optional)
                maximum number of iterations
            **kwargs:
                arguments passed to Resp.get_constraint_matrices

            Returns
            -------
            charges: ndarray
            """
            return super(cls, self).fit(hyp_a=hyp_a, hyp_b=hyp_b, restraint=restraint, basis=basis,
                                        method=method, ihfree=ihfree, **kwargs)

        @clean_kwargs
        def iter_solve(self, q, a, b, hyp_a=0.0005, tol=1e-6, maxiter=50, **kwargs):
            """
            Fit the charges iteratively, as required for the hyperbola penalty
            function.

            Parameters
            ----------
            q: ndarray
                partial atomic charges
            a: ndarray
                unrestrained matrix A
            b: ndarray
                matrix B
            hyp_a: float (optional)
                scale factor of asymptote limits of hyperbola
            tol: float (optional)
                threshold for convergence
            maxiter: int (optional)
                maximum number of iterations

            Returns
            -------
            charges: ndarray
            """

            return super(cls, self).iter_solve(q, a, b, hyp_a=hyp_a,
                                               hyp_b=hyp_b,
                                               ihfree=ihfree,
                                               tol=tol,
                                               maxiter=maxiter)

        @clean_kwargs
        def get_constraint_matrices(self, chrconstr=[], chrequiv=[],
                                    weights=1,
                                    vdw_scale_factors=(1.4, 1.6, 1.8, 2.0),
                                    vdw_point_density=1.0, vdw_radii={},
                                    rmin=0, rmax=-1,
                                    psi4_options={}, save_files=False, **kwargs):
            """
            Get A and B matrices to solve for charges, including charge constraints.

            Parameters
            ----------
            chrconstr: list or dict (optional)
                Intramolecular charge constraints in the form of
                {charge: atom_number_list} or [(charge, atom_number_list)].
                The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
                mean that atoms 1 and 2 together have a charge of 0.
            chrequiv: list (optional)
                lists of atoms with equivalent charges.
                e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
                charges, and atoms 3, 4, and 5 have equal charges.
            weights: iterable of ints (optional)
                weights of each conformer
            vdw_scale_factors: iterable of floats (optional)
                scale factors
            vdw_point_density: float (optional)
                point density
            use_radii: str (optional)
                which set of van der Waals' radii to use
            vdw_radii: dict (optional)
                van der Waals' radii. If elements in the molecule are not
                defined in the chosen ``use_radii`` set, they must be given here.
            rmin: float (optional)
                inner boundary of shell to keep grid points from
            rmax: float (optional)
                outer boundary of shell to keep grid points from. If < 0,
                all points are selected.
            psi4_options: dict (optional)
                additional Psi4 options
            save_files: bool (optional)
                if ``True``, Psi4 files are saved and the computed Esp
                and grids are written to files.

            Returns
            -------
            a: ndarray
            b: ndarray
            """
            return super(cls, self).get_constraint_matrices(method=method, basis=basis,
                                                            chrconstr=chrconstr,
                                                            chrequiv=chrequiv,
                                                            weights=weights, use_radii=use_radii,
                                                            vdw_point_density=vdw_point_density,
                                                            vdw_scale_factors=vdw_scale_factors,
                                                            vdw_radii=vdw_radii, rmin=rmin,
                                                            rmax=rmax, solvent=solvent,
                                                            psi4_options=psi4_options,
                                                            save_files=save_files)
        cls.run = run
        cls.optimize_geometry = optimize_geometry
        cls.fit = fit
        cls.iter_solve = iter_solve
        cls.get_constraint_matrices = get_constraint_matrices
        return cls
    return wrapper


@due.dcite(
    Doi('10.1021/j100142a004'),
    description='RESP-A1 model',
    path='psiresp.configs',
)
@resp_config(stage_2=True, hyp_a1=0.0005, hyp_a2=0.001, hyp_b=0.1,
             restraint=True, basis='6-31g*', method='scf',
             ihfree=True, use_radii='msk', solvent=None)
class RespA1(Resp):
    """
    Class to manage a 2-stage RESP fit for one 
    molecule of multiple conformers. This is called the RESP-A1 
    model in R.E.D.

    Please cite [Bayly1993]_ if you use this class.

    Parameters
    ----------
    conformers: iterable of Conformers
        conformers of the molecule. They must all have the same atoms
        in the same order.
    name: str (optional)
        name of the molecule. This is used to name output files. If not
        given, defaults to 'Mol'.
    chrconstr: list or dict (optional)
        Intramolecular charge constraints in the form of
        {charge: atom_number_list} or [(charge, atom_number_list)].
        The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
        mean that atoms 1 and 2 together have a charge of 0.
    chrequiv: list (optional)
        lists of atoms with equivalent charges, indexed from 1.
        e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
        charges, and atoms 3, 4, and 5 have equal charges.

    Attributes
    ----------
    name: str
        name of the molecule. This is used to name output files.
    n_atoms: int
        number of atoms in each conformer
    indices: ndarray of ints
        indices of each atom, indexed from 0
    heavy_ids: ndarray of ints
        atom numbers of heavy atoms, indexed from 1
    h_ids: ndarray of ints
        atom numbers of hydrogen atoms, indexed from 1
    symbols: ndarray
        element symbols
    charge: int or float
        overall charge of molecule
    conformers: list of Conformers
    n_conf: int
        number of conformers
    unrestrained_charges: ndarray of floats
        partial atomic charges (only exists after calling fit or run)
    restrained_charges: ndarray of floats
        partial atomic charges (only exists after calling fit or run)
    """


@due.dcite(
    Doi('10.1039/c0cp00111b'),
    description='RESP-A2 model',
    path='psiresp.configs',
)
@resp_config(stage_2=False, hyp_a1=0.01, hyp_a2=0.0, hyp_b=0.1,
             restraint=True, basis='6-31g*', method='scf',
             ihfree=True, use_radii='msk', solvent=None)
class RespA2(Resp):
    """
    Class to manage a 1-stage RESP fit for one 
    molecule of multiple conformers. This is called the RESP-A2
    model in R.E.D.

    Please cite [Dupradeau2010]_ if you use this class.

    Parameters
    ----------
    conformers: iterable of Conformers
        conformers of the molecule. They must all have the same atoms
        in the same order.
    name: str (optional)
        name of the molecule. This is used to name output files. If not
        given, defaults to 'Mol'.
    chrconstr: list or dict (optional)
        Intramolecular charge constraints in the form of
        {charge: atom_number_list} or [(charge, atom_number_list)].
        The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
        mean that atoms 1 and 2 together have a charge of 0.
    chrequiv: list (optional)
        lists of atoms with equivalent charges, indexed from 1.
        e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
        charges, and atoms 3, 4, and 5 have equal charges.

    Attributes
    ----------
    name: str
        name of the molecule. This is used to name output files.
    n_atoms: int
        number of atoms in each conformer
    indices: ndarray of ints
        indices of each atom, indexed from 0
    heavy_ids: ndarray of ints
        atom numbers of heavy atoms, indexed from 1
    h_ids: ndarray of ints
        atom numbers of hydrogen atoms, indexed from 1
    symbols: ndarray
        element symbols
    charge: int or float
        overall charge of molecule
    conformers: list of Conformers
    n_conf: int
        number of conformers
    unrestrained_charges: ndarray of floats
        partial atomic charges (only exists after calling fit or run)
    restrained_charges: ndarray of floats
        partial atomic charges (only exists after calling fit or run)
    """


@due.dcite(
    Doi('10.1002/jcc.540050204'),
    description='ESP-A1 model',
    path='psiresp.configs',
)
@resp_config(stage_2=False, hyp_a1=0.0, hyp_a2=0.0, hyp_b=0.1,
             restraint=False, basis='6-31g*', method='scf',
             ihfree=True, use_radii='msk', solvent=None)
class EspA1(Resp):
    """
    Class to manage a 1-stage unrestrained fit for one 
    molecule of multiple conformers. This is called the ESP-A1
    model in R.E.D.

    Please cite [Singh1984]_ if you use this class.

    Parameters
    ----------
    conformers: iterable of Conformers
        conformers of the molecule. They must all have the same atoms
        in the same order.
    name: str (optional)
        name of the molecule. This is used to name output files. If not
        given, defaults to 'Mol'.
    chrconstr: list or dict (optional)
        Intramolecular charge constraints in the form of
        {charge: atom_number_list} or [(charge, atom_number_list)].
        The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
        mean that atoms 1 and 2 together have a charge of 0.
    chrequiv: list (optional)
        lists of atoms with equivalent charges, indexed from 1.
        e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
        charges, and atoms 3, 4, and 5 have equal charges.

    Attributes
    ----------
    name: str
        name of the molecule. This is used to name output files.
    n_atoms: int
        number of atoms in each conformer
    indices: ndarray of ints
        indices of each atom, indexed from 0
    heavy_ids: ndarray of ints
        atom numbers of heavy atoms, indexed from 1
    h_ids: ndarray of ints
        atom numbers of hydrogen atoms, indexed from 1
    symbols: ndarray
        element symbols
    charge: int or float
        overall charge of molecule
    conformers: list of Conformers
    n_conf: int
        number of conformers
    unrestrained_charges: ndarray of floats
        partial atomic charges (only exists after calling fit or run)
    """


@due.dcite(
    Doi('10.1002/jcc.540050204'),
    Doi('10.1039/c0cp00111b'),
    description='ESP-A2 model',
    path='psiresp.configs',
)
@resp_config(stage_2=False, hyp_a1=0.0, hyp_a2=0.0, hyp_b=0.1,
             restraint=False, basis='sto-3g', method='scf',
             ihfree=True, use_radii='msk', solvent=None)
class EspA2(Resp):
    """
    Class to manage a 1-stage unrestrained fit for one 
    molecule of multiple conformers, using HF/STO-3G. 
    This is called the ESP-A2 model in R.E.D.

    Please cite [Singh1984]_ and [Dupradeau2010]_ if you use this class.

    Parameters
    ----------
    conformers: iterable of Conformers
        conformers of the molecule. They must all have the same atoms
        in the same order.
    name: str (optional)
        name of the molecule. This is used to name output files. If not
        given, defaults to 'Mol'.
    chrconstr: list or dict (optional)
        Intramolecular charge constraints in the form of
        {charge: atom_number_list} or [(charge, atom_number_list)].
        The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
        mean that atoms 1 and 2 together have a charge of 0.
    chrequiv: list (optional)
        lists of atoms with equivalent charges, indexed from 1.
        e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
        charges, and atoms 3, 4, and 5 have equal charges.

    Attributes
    ----------
    name: str
        name of the molecule. This is used to name output files.
    n_atoms: int
        number of atoms in each conformer
    indices: ndarray of ints
        indices of each atom, indexed from 0
    heavy_ids: ndarray of ints
        atom numbers of heavy atoms, indexed from 1
    h_ids: ndarray of ints
        atom numbers of hydrogen atoms, indexed from 1
    symbols: ndarray
        element symbols
    charge: int or float
        overall charge of molecule
    conformers: list of Conformers
    n_conf: int
        number of conformers
    unrestrained_charges: ndarray of floats
        partial atomic charges (only exists after calling fit or run)
    """


@due.dcite(
    Doi('10.1021/ct200196m'),
    description='ATB model',
    path='psiresp.configs',
)
#: TODO: autodetect symmetry?
@resp_config(stage_2=False, hyp_a1=0.0, hyp_a2=0.0, hyp_b=0.1,
             restraint=False, basis='6-31g*', method='b3lyp',
             ihfree=False, use_radii='msk', solvent='water')
class ATBResp(Resp):
    """
    Class to manage a 1-stage unrestrained fit using the method of 
    Singh and Kollman 1984 at B3LYP/6-31G* with implicit solvent, as 
    described for the Automated Topology Builder.

    Please cite [Malde2011]_ and [Singh1984]_ if you use this class.

    Parameters
    ----------
    conformers: iterable of Conformers
        conformers of the molecule. They must all have the same atoms
        in the same order.
    name: str (optional)
        name of the molecule. This is used to name output files. If not
        given, defaults to 'Mol'.
    chrconstr: list or dict (optional)
        Intramolecular charge constraints in the form of
        {charge: atom_number_list} or [(charge, atom_number_list)].
        The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]]
        mean that atoms 1 and 2 together have a charge of 0.
    chrequiv: list (optional)
        lists of atoms with equivalent charges, indexed from 1.
        e.g. [[1, 2], [3, 4, 5]] mean that atoms 1 and 2 have equal
        charges, and atoms 3, 4, and 5 have equal charges.

    Attributes
    ----------
    name: str
        name of the molecule. This is used to name output files.
    n_atoms: int
        number of atoms in each conformer
    indices: ndarray of ints
        indices of each atom, indexed from 0
    heavy_ids: ndarray of ints
        atom numbers of heavy atoms, indexed from 1
    h_ids: ndarray of ints
        atom numbers of hydrogen atoms, indexed from 1
    symbols: ndarray
        element symbols
    charge: int or float
        overall charge of molecule
    conformers: list of Conformers
    n_conf: int
        number of conformers
    unrestrained_charges: ndarray of floats
        partial atomic charges (only exists after calling fit or run)
    """


def multiresp_config(stage_2=True, hyp_a1=0.0005, hyp_a2=0.001, hyp_b=0.1,
                     restraint=True, basis='6-31g*', method='scf', use_radii='msk',
                     ihfree=True, solvent=None):
    """Make analogous classes to R.E.D. options"""
    def wrapper(cls):
        @clean_kwargs
        def run(self, opt=False, save_opt_geometry=False,
                intra_chrconstr=[], intra_chrequiv=[], inter_chrconstr=[],
                inter_chrequiv=[], n_orient=0, orient=[], n_rotate=0, rotate=[],
                n_translate=0, translate=[], equal_methyls=False, psi4_options={},
                **kwargs):
            """
            Parameters
            ----------
            opt: bool (optional)
                Whether to optimise the geometry of each conformer
            save_opt_geometry: bool (optional)
                if ``True``, writes optimised geometries to an XYZ file
            intra_chrconstr: list of lists or dicts (optional)
                Intramolecular charge constraints for each molecule in the 
                form of {charge: atom_number_list} or [(charge, atom_number_list)]. 
                The numbers are indexed from 1. e.g. {0: [1, 2]} or [[0, [1, 2]]] 
                mean that atoms 1 and 2 together have a charge of 0.
            intra_chrequiv: list of lists (optional)
                Lists of atoms with equivalent charges within each molecule.
                e.g. [
                    [[1, 2], [3, 4, 5]],
                    [[1, 3, 5, 7]]
                    ] mean that atoms 1 and 2 in the first molecule have equal 
                charges; atoms 3, 4, and 5 in the first molecule have 
                equal charges; atoms 1, 3, 5, 7 in the second molecule have equal 
                charges.
            inter_chrconstr: list of lists (optional)
                Intermolecular charge constraints in the form of 
                {charge: [(mol_number, atom_number)]} or 
                [(charge, [(mol_number, atom_number)])]. 
                The numbers are indexed from 1.
                e.g. {0: [(1, 3), (2, 1)]} or [(0, [(1, 3), (2, 1)])] mean that
                the third atom of the first molecule, and the first atom of the 
                second molecule, combine to have a charge of 0.
            inter_chrequiv: list of lists (optional)
                Lists of atoms with equivalent charges between each molecule, 
                in the form [(mol_number, atom_number)]. 
                e.g. [[(1, 2), (2, 2), (3, 4)]] mean that atom 2 of molecule 1, 
                atom 2 of molecule 2, and atom 4 of molecule 3, all have equal 
                charges.
            n_orient: int (optional)
                If this is greater than 0 and ``orient`` is not given, 
                ``n_orient`` orientations are automatically generated for each 
                molcule. Heavy atoms are prioritised.
            orient: list of lists (optional)
                List of lists of reorientations. 
                Corresponds to REMARK REORIENT in R.E.D.
                e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
                around the first, fifth and ninth atom; and the second in reverse 
                order.
            n_rotate: int (optional)
                If this is greater than 0 and ``rotate`` is not given, 
                ``n_rotate`` rotations are automatically generated for each 
                molecule. Heavy atoms are prioritised.
            rotate: list of lists (optional)
                List of lists of rotations. 
                Corresponds to REMARK ROTATE in R.E.D.
                e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
                around the first, fifth and ninth atom; and the second in reverse 
                order.
            n_translate: int (optional)
                If this is greater than 0 and ``translate`` is not given, 
                ``n_translate`` translations are randomly generated for each 
                molcule in the domain [0, 1).
            translate: list of lists (optional)
                List of lists of translations. 
                Corresponds to REMARK TRANSLATE in R.E.D.
                e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the 
                x coordinates, 0 to the y coordinates, and -0.5 to the z 
                coordinates.
            psi4_options: dict (optional)
                additional Psi4 options
            **kwargs:
                arguments passed to MultiResp.fit

            Returns
            -------
            charges: list of ndarrays
            """
            super(cls, self).run(opt=opt, save_opt_geometry=save_opt_geometry,
                                 intra_chrconstr=intra_chrconstr,
                                 intra_chrequiv=intra_chrequiv,
                                 inter_chrconstr=inter_chrconstr,
                                 inter_chrequiv=inter_chrequiv,
                                 psi4_options=psi4_options,
                                 n_orient=n_orient, orient=orient,
                                 n_rotate=n_rotate, rotate=rotate,
                                 n_translate=n_translate, translate=translate,
                                 equal_methyls=equal_methyls, stage_2=stage_2,
                                 hyp_a1=hyp_a1, hyp_a2=hyp_a2, hyp_b=hyp_b,
                                 restraint=restraint, basis=basis, solvent=solvent,
                                 method=method, ihfree=ihfree, **kwargs)

        def optimize_geometry(self, psi4_options={}, save_opt_geometry=False):
            """
            Optimise geometry for all conformers.

            Parameters
            ----------
            psi4_options: dict (optional)
                additional Psi4 options
            save_opt_geometry: bool (optional)
                if ``True``, saves optimised geometries to XYZ file
            """
            super(cls, self).optimize_geometry(method=method, basis=basis,
                                               psi4_options=psi4_options,
                                               save_opt_geometry=save_opt_geometry)

        @clean_kwargs
        def fit(self, tol=1e-6, maxiter=50, **kwargs):
            """
            Perform the Resp fits.

            Parameters
            ----------
            tol: float (optional)
                threshold for convergence
            maxiter: int (optional)
                maximum number of iterations
            **kwargs:
                arguments passed to Resp.get_constraint_matrices

            Returns
            -------
            charges: ndarray
            """
            super(cls, self).fit(hyp_b=hyp_b, restraint=restraint, basis=basis,
                                 method=method, ihfree=ihfree, **kwargs)

        @clean_kwargs
        def get_constraint_matrices(self, intra_chrconstr=[], intra_chrequiv=[],
                                    inter_chrconstr=[], inter_chrequiv=[],
                                    weights=1, **kwargs):
            """
            Get A and B matrices to solve for charges, including charge constraints.

            Parameters
            ----------
            intra_chrconstr: list of lists or dicts (optional)
                Intramolecular charge constraints for each molecule in the 
                form of, where the list of constraints has the form 
                [{charge: atom_number_list}] or [[(charge, atom_number_list)]]. 
                The numbers are indexed from 1. 
                e.g. [{0: [1, 2]}, {0.2: [3, 4, 5]}]
                mean that atoms 1 and 2 together have a charge of 0 in the first
                molecule, whereas atoms 3, 4, and 5 combine to a charge of 0.2 in 
                the second molecule.
            intra_chrequiv: list of lists (optional)
                Lists of atoms with equivalent charges within each molecule.
                e.g. [
                    [[1, 2], [3, 4, 5]],
                    [[1, 3, 5, 7]]
                    ] mean that atoms 1 and 2 in the first molecule have equal 
                charges; atoms 3, 4, and 5 in the first molecule have 
                equal charges; atoms 1, 3, 5, 7 in the second molecule have equal 
                charges.
            inter_chrconstr: list of lists (optional)
                Intermolecular charge constraints in the form of 
                {charge: [(mol_number, atom_number)]} or 
                [(charge, [(mol_number, atom_number)])]. 
                The numbers are indexed from 1.
                e.g. {0: [(1, 3), (2, 1)]} or [(0, [(1, 3), (2, 1)])] mean that
                the third atom of the first molecule, and the first atom of the 
                second molecule, combine to have a charge of 0.
            inter_chrequiv: list of lists (optional)
                Lists of atoms with equivalent charges between each molecule, 
                in the form [(mol_number, atom_number)]. 
                e.g. [[(1, 2), (2, 2), (3, 4)]] mean that atom 2 of molecule 1, 
                atom 2 of molecule 2, and atom 4 of molecule 3, all have equal 
                charges.
            weights: iterable (optional)
                List of weights for each molecule, or list of lists of weights for
                each conformer. e.g. [1, 2, [3, 4]] weights all conformers in the 
                first molecule by 1; all the conformers in the second molecule by 2;
                the first conformer in the third molecule by 3; and the second 
                conformer in the third molecule by 4. If only one value is given, it
                is repeated for every molecule.
            **kwargs
                Arguments passed to Resp.get_constraint_matrices.

            Returns
            -------
            A: ndarray
            B: ndarray
            edges: list of tuples
                list of start, end indices of the molecule atoms in `a` and `b`.
                For example, ``i, j = edges[3]`` such that ``a[i:j, i:j]`` is 
                the A matrix for the 4th molecule.
            nmol: ndarray
                array containing the number of orientations in each Resp, with 
                the same shape as B
            """

            super(cls, self).get_constraint_matrices(intra_chrconstr=intra_chrconstr,
                                                     intra_chrequiv=intra_chrequiv,
                                                     inter_chrconstr=inter_chrconstr,
                                                     inter_chrequiv=inter_chrequiv,
                                                     solvent=solvent, weights=weights,
                                                     method=method, basis=basis,
                                                     use_radii=use_radii,
                                                     **kwargs)
            cls.run = run
            cls.optimize_geometry = optimize_geometry
            cls.fit = fit
            cls.get_constraint_matrices = get_constraint_matrices
        return cls
    return wrapper


@due.dcite(
    Doi('10.1021/j100142a004'),
    description='RESP-A1 multi-molecule model',
    path='psiresp.configs',
)
@multiresp_config(stage_2=True, hyp_a1=0.0005, hyp_a2=0.001, hyp_b=0.1,
                  restraint=True, basis='6-31g*', method='scf',
                  ihfree=True, use_radii='msk', solvent=None)
class MultiRespA1(MultiResp):
    """
    Class to manage a 2-stage RESP fit for multiple molecules of 
    multiple conformers. This is called the RESP-A1 model in R.E.D.

    Please cite [Bayly1993]_ if you use this class.

    Parameters
    ----------
    resps: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.

    Attributes
    ----------
    molecules: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.
    n_molecules: int
        number of molecule Resp classes
    n_structures: int
        number of structures in entire MultiResp fit, where one structure 
        is one orientation of one conformer
    charges: list of ndarray
        partial atomic charges for each molecule
        (only exists after calling fit or run)
    """


@due.dcite(
    Doi('10.1039/c0cp00111b'),
    description='RESP-A2 multi-molecule model',
    path='psiresp.configs',
)
@multiresp_config(stage_2=False, hyp_a1=0.01, hyp_a2=0.0, hyp_b=0.1,
                  restraint=True, basis='6-31g*', method='scf',
                  ihfree=True, use_radii='msk', solvent=None)
class MultiRespA2(MultiResp):
    """
    Class to manage a 1-stage RESP fit for multiple molecules of 
    multiple conformers. This is called the RESP-A2 
    model in R.E.D.

    Please cite [Dupradeau2010]_ if you use this class.

    Parameters
    ----------
    resps: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.

    Attributes
    ----------
    molecules: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.
    n_molecules: int
        number of molecule Resp classes
    n_structures: int
        number of structures in entire MultiResp fit, where one structure 
        is one orientation of one conformer
    charges: list of ndarray
        partial atomic charges for each molecule
        (only exists after calling fit or run)
    """


@due.dcite(
    Doi('10.1002/jcc.540050204'),
    description='ESP-A1 multi-molecule model',
    path='psiresp.configs',
)
@multiresp_config(stage_2=False, hyp_a1=0.0, hyp_a2=0.0, hyp_b=0.1,
                  restraint=True, basis='6-31g*', method='scf',
                  ihfree=True, use_radii='msk', solvent=None)
class MultiEspA1(MultiResp):
    """
    Class to manage a 1-stage unrestrained fit for multiple molecules of 
    multiple conformers. This is called the ESP-A1 model in R.E.D.

    Please cite [Singh1984]_ if you use this class.

    Parameters
    ----------
    resps: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.

    Attributes
    ----------
    molecules: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.
    n_molecules: int
        number of molecule Resp classes
    n_structures: int
        number of structures in entire MultiResp fit, where one structure 
        is one orientation of one conformer
    charges: list of ndarray
        partial atomic charges for each molecule
        (only exists after calling fit or run)
    """


@due.dcite(
    Doi('10.1002/jcc.540050204'),
    Doi('10.1039/c0cp00111b'),
    description='ESP-A2 multi-molecule model',
    path='psiresp.configs',
)
@multiresp_config(stage_2=False, hyp_a1=0.0, hyp_a2=0.0, hyp_b=0.1,
                  restraint=True, basis='sto-3g', method='scf',
                  ihfree=True, use_radii='msk', solvent=None)
class MultiEspA2(MultiResp):
    """
    Class to manage a 1-stage unrestrained fit for multiple molecules of 
    multiple conformers. This is called the ESP-A2 model in R.E.D.

    Please cite [Singh1984]_ and [Dupradeau2010]_ if you use this class.

    Parameters
    ----------
    resps: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.

    Attributes
    ----------
    molecules: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.
    n_molecules: int
        number of molecule Resp classes
    n_structures: int
        number of structures in entire MultiResp fit, where one structure 
        is one orientation of one conformer
    charges: list of ndarray
        partial atomic charges for each molecule
        (only exists after calling fit or run)
    """


@due.dcite(
    Doi('10.1021/ct200196m'),
    description='ATB multi-molecule model',
    path='psiresp.configs',
)
#: TODO: autodetect symmetry?
@multiresp_config(stage_2=False, hyp_a1=0.0, hyp_a2=0.0, hyp_b=0.1,
                  restraint=False, basis='6-31g*', method='b3lyp',
                  ihfree=False, use_radii='msk', solvent='water')
class ATBMultiResp(MultiResp):
    """
    Class to manage a 1-stage unrestrained fit using the method of 
    Singh and Kollman 1984 at B3LYP/6-31G* with implicit solvent, as 
    described for the Automated Topology Builder.

    Please cite [Malde2011]_ and [Singh1984]_ if you use this class.

    Parameters
    ----------
    resps: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.

    Attributes
    ----------
    molecules: list of Resp
        Molecules for multi-molecule fit, set up in Resp classes.
    n_molecules: int
        number of molecule Resp classes
    n_structures: int
        number of structures in entire MultiResp fit, where one structure 
        is one orientation of one conformer
    charges: list of ndarray
        partial atomic charges for each molecule
        (only exists after calling fit or run)
    """



from __future__ import absolute_import
import logging
import os
import tempfile

import numpy as np

from . import utils
from .orientation import Orientation

log = logging.getLogger(__name__)


class Conformer(object):
    """
    Wrapper class to manage one conformer molecule, containing 
    multiple orientations.

    Parameters
    ----------
    molecule: Psi4 molecule
    charge: int (optional)
        overall charge of the molecule.
    multiplicity: int (optional)
        multiplicity of the molecule
    name: str (optional)
        name of the molecule. This is used to name output files. If not 
        given, the default Psi4 name is 'default'.
    orient: list of tuples of ints (optional)
        List of reorientations. Corresponds to REMARK REORIENT in R.E.D.
        e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
        around the first, fifth and ninth atom; and the second in reverse 
        order.
    rotate: list of tuples of ints (optional)
        List of rotations. Corresponds to REMARK ROTATE in R.E.D.
        e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
        around the first, fifth and ninth atom; and the second in reverse 
        order.
    translate: list of tuples of floats (optional)
        List of translations. Corresponds to REMARK TRANSLATE in R.E.D.
        e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the 
        x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
    grid_name: str (optional)
            template for grid filename for each Orientation.
    esp_name: str (optional)
        template for ESP filename for each Orientation.
    load_files: bool (optional)
        If ``True``, tries to load ESP and grid data from file.

    Attributes
    ----------
    molecule: Psi4 molecule
    name: str
        name of the molecule. This is used to name output files.
    n_atoms: int
        number of atoms in each conformer
    symbols: ndarray
        element symbols
    orientations: list of Orientations
        list of the molecule with reoriented coordinates
    """

    def __init__(self, molecule, charge=0, multiplicity=1, name=None,
                 orient=[], rotate=[], translate=[], load_files=False,
                 grid_name='grid.dat', esp_name='grid_esp.dat'):
        if name and name != molecule.name():
            molecule.set_name(name)
        self.name = molecule.name()
        if charge != molecule.molecular_charge():
            molecule.set_molecular_charge(charge)
        if multiplicity != molecule.multiplicity():
            molecule.set_multiplicity(multiplicity)
        self.charge = charge
        self.multiplicity = multiplicity
        self.molecule = molecule
        self.n_atoms = molecule.natom()
        self.symbols = np.array([molecule.symbol(i) for i in range(self.n_atoms)])
        self._orient = orient[:]
        self._rotate = rotate[:]
        self._translate = translate[:]
        self._load_files = load_files
        self._grid_name = grid_name
        self._esp_name = esp_name
        self._orientations = []
        self._orientation = Orientation(self.molecule, symbols=self.symbols)

    @property
    def n_orientations(self):
        return len(self.orientations)

    @property
    def orientations(self):
        if not self._orientations:
            self._gen_orientations(orient=self._orient,
                                   translate=self._translate,
                                   rotate=self._rotate,
                                   load_files=self._load_files)
        if not self._orientations:
            return [self._orientation]
        return self._orientations

    @property
    def _grid_needs_computing(self):
        return any(mol.grid is None for mol in self.orientations)

    @property
    def coordinates(self):
        return self.molecule.geometry().np.astype('float')

    def clone(self, name=None):
        """Clone into another instance of Conformer

        Parameters
        ----------
        name: str (optional)
            If not given, the new Conformer instance has '_copy' appended
            to its name

        Returns
        -------
        Conformer
        """
        if name is None:
            name = self.name+'_copy'

        new = type(self)(self.molecule.clone(), name=name,
                         charge=self.charge,
                         multiplicity=self.multiplicity,
                         orient=self._orient, rotate=self._rotate,
                         translate=self._translate,
                         load_files=self._load_files,
                         grid_name=self._grid_name,
                         esp_name=self._esp_name)
        return new

    def _gen_orientations(self, orient=[], translate=[], rotate=[],
                          load_files=False):
        """
        Generate new orientations.

        Parameters
        ----------
        orient: list of tuples of ints (optional)
            List of reorientations. Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse 
            order.
        rotate: list of tuples of ints (optional)
            List of rotations. Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse 
            order.
        translate: list of tuples of floats (optional)
            List of translations. Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the 
            x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
        load_files: bool (optional)
            If ``True``, each orientation tries to load ESP and grid data from file.
        """

        for atom_ids in orient:
            a, b, c = [a-1 if a > 0 else a for a in atom_ids]
            xyz = utils.orient_rigid(a, b, c, self.coordinates)
            self._add_orientation(xyz, load_files=load_files)

        for atom_ids in rotate:
            a, b, c = [a-1 if a > 0 else a for a in atom_ids]
            xyz = utils.rotate_rigid(a, b, c, self.coordinates)
            self._add_orientation(xyz, load_files=load_files)

        for translation in translate:
            xyz = self.coordinates+translation
            self._add_orientation(xyz, load_files=load_files)

    def _add_orientation(self, coordinates, load_files=False):
        import psi4
        mat = psi4.core.Matrix.from_array(coordinates)
        cmol = self.molecule.clone()
        cmol.set_geometry(mat)
        cmol.fix_com(True)
        cmol.fix_orientation(True)
        cmol.update_geometry()
        cmol.set_name('{}_o{}'.format(self.name, len(self._orientations)+1))
        self._orientations.append(Orientation(cmol, symbols=self.symbols,
                                              load_files=load_files,
                                              grid_name=self._grid_name,
                                              esp_name=self._esp_name))

    def add_orientations(self, orient=[], translate=[], rotate=[],
                         load_files=False):
        """
        Add new orientations to generate.

        Parameters
        ----------
        orient: list of tuples of ints (optional)
            List of reorientations. Corresponds to REMARK REORIENT in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two reorientations: the first
            around the first, fifth and ninth atom; and the second in reverse 
            order.
        rotate: list of tuples of ints (optional)
            List of rotations. Corresponds to REMARK ROTATE in R.E.D.
            e.g. [(1, 5, 9), (9, 5, 1)] creates two rotations: the first
            around the first, fifth and ninth atom; and the second in reverse 
            order.
        translate: list of tuples of floats (optional)
            List of translations. Corresponds to REMARK TRANSLATE in R.E.D.
            e.g. [(1.0, 0, -0.5)] creates a translation that adds 1.0 to the 
            x coordinates, 0 to the y coordinates, and -0.5 to the z coordinates.
        load_files: bool (optional)
            If ``True``, each orientation tries to load ESP and grid data from file.
        """
        self._orient.extend(orient)
        self._translate.extend(translate)
        self._rotate.extend(rotate)
        if self._orientations:
            self._gen_orientations(orient=orient, rotate=rotate,
                                   translate=translate, load_files=load_files)

    def optimize_geometry(self, method='scf', basis='6-31g*',
                          psi4_options={}, save_opt_geometry=True,
                          save_files=False):
        """
        Optimise the geometry of the molecule and update the coordinates.

        Parameters
        ----------
        multiplicity: int (optional)
        basis: str (optional)
            Basis set to optimise geometry
        method: str (optional)
            Method to optimise geometry
        psi4_options: dict (optional)
            additional Psi4 options
        save_opt_geometry: bool (optional)
            if ``True``, writes the optimised geometry to an XYZ file.
        save_files: bool (optional)
            if ``True``, Psi4 files are saved. This does not affect 
            writing optimised geometries to files. 
        """
        if not save_files:
            cwd = os.getcwd()
            with tempfile.TemporaryDirectory(prefix='tmp') as tmpdir:
                os.chdir(tmpdir)
                self._optimize_geometry(method=method, basis=basis,
                                        psi4_options=psi4_options)
                os.chdir(cwd)
        else:
            self._optimize_geometry(method=method, basis=basis,
                                    psi4_options=psi4_options)

        if save_opt_geometry:
            self.molecule.save_xyz_file(self.name+'_opt.xyz', True)

    def _optimize_geometry(self, method='scf', basis='6-31g*', psi4_options={}):
        import psi4
        psi4.set_options({'basis': basis,
                          'geom_maxiter': 200,
                          'full_hess_every': 10,
                          'g_convergence': 'gau_tight',
                          })
        psi4.set_options(psi4_options)
        logfile = self.name+'_opt.log'
        psi4.set_output_file(logfile, False)  # doesn't work? where is the output?!
        psi4.optimize(method, molecule=self.molecule)
        psi4.core.clean()
        self._orientations = []

    def get_esp_matrices(self, weight=1.0, vdw_points=None, rmin=0, rmax=-1,
                         basis='6-31g*', method='scf', solvent=None,
                         psi4_options={}, save_files=False, vdw_radii={},
                         use_radii='msk', vdw_point_density=1.0,
                         vdw_scale_factors=(1.4, 1.6, 1.8, 2.0)):
        """
        Get A and B matrices to solve for charges, from Bayly:93:Eq.11:: Aq=B

        Parameters
        ----------
        weight: float (optional)
            how much to weight the matrices
        vdw_points: list of tuples (optional)
            List of tuples of (unit_shell_coordinates, scaled_atom_radii).
            If this is ``None`` or empty, new points are generated from 
            utils.gen_connolly_shells and the variables ``vdw_radii``,
            ``use_radii``, ``vdw_scale_factors``, ``vdw_point_density``.
        rmin: float (optional)
            inner boundary of shell to keep grid points from
        rmax: float (optional)
            outer boundary of shell to keep grid points from. If < 0,
            all points are selected.
        basis: str (optional)
            Basis set to compute ESP
        method: str (optional)
            Method to compute ESP
        solvent: str (optional)
            Solvent for computing in implicit solvent
        use_radii: str (optional)
            which set of van der Waals' radii to use. 
            Ignored if ``vdw_points`` is provided.
        vdw_scale_factors: iterable of floats (optional)
            scale factors. Ignored if ``vdw_points`` is provided.
        vdw_point_density: float (optional)
            point density. Ignored if ``vdw_points`` is provided.
        vdw_radii: dict (optional)
            van der Waals' radii. If elements in the molecule are not
            defined in the chosen ``use_radii`` set, they must be given here.
            Ignored if ``vdw_points`` is provided.
        psi4_options: dict (optional)
            additional Psi4 options
        save_files: bool (optional)
            if ``True``, Psi4 files are saved and the computed ESP
            and grids are written to files.

        Returns
        -------
        A: ndarray (n_atoms, n_atoms)
        B: ndarray (n_atoms,)
        """
        A = np.zeros((self.n_atoms, self.n_atoms))
        B = np.zeros(self.n_atoms)

        for mol in self.orientations:
            a, b = mol.get_esp_matrices(vdw_points=vdw_points,
                                        basis=basis, method=method,
                                        solvent=solvent,
                                        psi4_options=psi4_options,
                                        save_files=save_files)
            A += a
            B += b

        A *= (weight**2)
        B *= (weight**2)

        return A, B



from __future__ import division, absolute_import
import os
import logging
import tempfile
import textwrap
import warnings

import numpy as np

from . import utils

log = logging.getLogger(__name__)

#: convert bohr to angstrom
BOHR_TO_ANGSTROM = 0.52917721092


class Orientation(object):
    """
    Class to manage one Psi4 molecule.

    Parameters
    ----------
    molecule: Psi4 molecule
    symbols: list (optional)
        molecule elements. If not given, this is generated from 
        the Psi4 molecule.
    grid_name: str (optional)
        template for grid filename. If ``load_files=True``, the
        class tries to load a grid from $molname_$grid_name.
    esp_name: str (optional)
        template for ESP filename. If ``load_files=True``, the
        class tries to load ESP points from $molname_$esp_name.
    load_files: bool (optional)
        If ``True``, tries to load data from file.

    Attributes
    ----------
    molecule: Psi4 molecule
    symbols: ndarray
        molecule elements
    bohr: bool
        if the molecule coordinates are in bohr
    coordinates: ndarray
        molecule coordinates in angstrom
    grid_filename: str
        file to load grid data from, or save it to.
    esp_filename: str
        file to load ESP data from, or save it to.
    grid: ndarray
        grid of points to compute ESP for. Only populated when
        ``get_grid()`` is called.
    esp: ndarray
        ESP at grid points. Only populated when ``get_esp()``
        is called.
    r_inv: ndarray
        inverse distance from each grid point to each atom, in
        atomic units. Only populated when ``get_inverse_distance()``
        is called.
    """

    def __init__(self, molecule, symbols=None, grid_name='grid.dat',
                 esp_name='grid_esp.dat', load_files=False):
        self.name = molecule.name()
        self.n_atoms = molecule.natom()
        if symbols is None:
            symbols = np.array([molecule.symbol(i) for i in range(self.n_atoms)])
        self.symbols = symbols
        self.bohr = 'Bohr' in str(molecule.units())
        self.indices = np.arange(self.n_atoms).astype(int)
        self.molecule = molecule
        self._grid_name = grid_name
        self._esp_name = esp_name
        self.grid_filename = self._prepend_name_to_file(grid_name)
        self.esp_filename = self._prepend_name_to_file(esp_name)
        self._grid = self._esp = self._r_inv = None

        # try to read from files
        if load_files:
            fail = 'Could not read data from {}'
            try:
                self.grid = np.loadtxt(self.grid_filename)
            except OSError:
                warnings.warn(fail.format(self.grid_filename))
            else:
                if self.bohr:
                    self.grid *= BOHR_TO_ANGSTROM
                log.info('Read grid from {}'.format(self.grid_filename))
            try:
                self.esp = np.loadtxt(self.esp_filename)
            except OSError:
                warnings.warn(fail.format(self.esp_filename))
            else:
                log.info('Read esp from {}'.format(self.esp_filename))

    def _prepend_name_to_file(self, filename):
        head, tail = os.path.split(filename)
        if head and not head.endswith(r'/'):
            head += '/'
        return '{}{}_{}'.format(head, self.name, tail)

    @property
    def grid(self):
        return self._grid

    @grid.setter
    def grid(self, values):
        # ESP and inverse distance depend on grid; redo
        self._grid = values
        self._esp = None
        self._r_inv = None

    @property
    def esp(self):
        return self._esp

    @esp.setter
    def esp(self, values):
        self._esp = values
        self._r_inv = None

    @property
    def r_inv(self):
        """Inverse distance from each grid point to each atom, in atomic units."""
        if self._r_inv is None and self.grid is not None:
            points = self.grid.reshape((len(self.grid), 1, 3))
            disp = self.coordinates-points
            inverse = 1/np.sqrt(np.einsum('ijk, ijk->ij', disp, disp))
            self._r_inv = inverse*BOHR_TO_ANGSTROM

        return self._r_inv

    @property
    def coordinates(self):
        coords = self.molecule.geometry().np.astype('float')*BOHR_TO_ANGSTROM
        return coords

    def clone(self, name=None):
        """Clone into another instance of Orientation

        Parameters
        ----------
        name: str (optional)
            If not given, the new Resp instance has the same name as this one

        Returns
        -------
        Orientation
        """
        mol = self.molecule.clone()
        if name is not None:
            mol.set_name(name)
        new = type(self)(name, symbols=self.symbols,
                         grid_name=self._grid_name,
                         esp_name=self._esp_name)
        return new

    def get_esp_matrices(self, vdw_points=None, rmin=0, rmax=-1,
                         basis='6-31g*', method='scf', solvent=None,
                         vdw_radii={}, use_radii='msk',
                         vdw_point_density=1.0,
                         vdw_scale_factors=(1.4, 1.6, 1.8, 2.0),
                         psi4_options={}, save_files=False):
        """
        Get A and B matrices to solve for charges, from Bayly:93:Eq.11:: Aq=B

        Parameters
        ----------
        vdw_points: list of tuples (optional)
            List of tuples of (unit_shell_coordinates, scaled_atom_radii).
            If this is ``None`` or empty, new points are generated from 
            utils.gen_connolly_shells and the variables ``vdw_radii``,
            ``use_radii``, ``vdw_scale_factors``, ``vdw_point_density``.
        use_radii: str (optional)
            which set of van der Waals' radii to use. 
            Ignored if ``vdw_points`` is provided.
        vdw_scale_factors: iterable of floats (optional)
            scale factors. Ignored if ``vdw_points`` is provided.
        vdw_point_density: float (optional)
            point density. Ignored if ``vdw_points`` is provided.
        vdw_radii: dict (optional)
            van der Waals' radii. If elements in the molecule are not
            defined in the chosen ``use_radii`` set, they must be given here.
            Ignored if ``vdw_points`` is provided.
        rmin: float (optional)
            inner boundary of shell to keep grid points from
        rmax: float (optional)
            outer boundary of shell to keep grid points from. If < 0,
            all points are selected.
        basis: str (optional)
            Basis set to compute ESP
        method: str (optional)
            Method to compute ESP
        solvent: str (optional)
            Solvent for computing in implicit solvent
        psi4_options: dict (optional)
            additional Psi4 options
        save_files: bool (optional)
            if ``True``, Psi4 files are saved and the computed ESP
            and grids are written to files.

        Returns
        -------
        a: ndarray
        b: ndarray
        """

        if self.grid is None:
            self.get_grid(vdw_points=vdw_points, rmin=rmin,
                          rmax=rmax, vdw_radii=vdw_radii,
                          use_radii=use_radii, density=vdw_point_density,
                          scale_factors=vdw_scale_factors,
                          save_files=save_files)

        if self.esp is None:
            self.get_esp(basis=basis, method=method,
                         psi4_options=psi4_options,
                         save_files=save_files, solvent=solvent)

        a = np.einsum('ij, ik->jk', self.r_inv, self.r_inv)
        b = np.einsum('i, ij->j', self.esp, self.r_inv)

        return a, b

    def get_grid(self, vdw_points=None, rmin=0, rmax=-1, fmt='%15.10f',
                 save_files=False, vdw_radii={}, use_radii='msk',
                 scale_factors=(1.4, 1.6, 1.8, 2.0),
                 density=1.0):
        """
        Get ESP grid from a given file or compute it from the
        van der Waals' surfaces.

        Parameters
        ----------
        vdw_points: list of tuples (optional)
            List of tuples of (unit_shell_coordinates, scaled_atom_radii).
            If this is ``None`` or empty, new points are generated from 
            utils.gen_connolly_shells and the variables ``vdw_radii``,
            ``use_radii``, ``scale_factors``, ``density``.
        rmin: float (optional)
            inner boundary of shell to keep grid points from
        rmax: float (optional)
            outer boundary of shell to keep grid points from. If < 0,
            all points are selected.
        use_radii: str (optional)
            which set of van der Waals' radii to use. 
            Ignored if ``vdw_points`` is provided.
        scale_factors: iterable of floats (optional)
            scale factors. Ignored if ``vdw_points`` is provided.
        density: float (optional)
            point density. Ignored if ``vdw_points`` is provided.
        vdw_radii: dict (optional)
            van der Waals' radii. If elements in the molecule are not
            defined in the chosen ``use_radii`` set, they must be given here.
            Ignored if ``vdw_points`` is provided.
        fmt: str (optional)
            float format
        save_files: bool (optional)
            if ``True``, Psi4 files are saved and the computed grid
            is written to a file.

        Returns
        -------
        grid: ndarray
        """
        # usually generated in RESP or Conformer
        if vdw_points is None or len(vdw_points) == 0:
            vdw_points = utils.gen_connolly_shells(self.symbols,
                                                   vdw_radii=vdw_radii,
                                                   use_radii=use_radii,
                                                   scale_factors=scale_factors,
                                                   density=density)
        points = []
        for pts, rad in vdw_points:
            points.append(utils.gen_vdw_surface(pts, rad, self.coordinates,
                                                rmin=rmin, rmax=rmax))
        points = to_save = np.concatenate(points)

        if save_files:
            if self.bohr:
                to_save = points/BOHR_TO_ANGSTROM
            np.savetxt(self.grid_filename, to_save, fmt=fmt)

        self.grid = points
        log.debug('Computed grid for {} with {} points'.format(self.name,
                                                               len(points)))
        return points

    def get_esp(self, basis='6-31g*', method='scf', solvent=None,
                psi4_options={}, fmt='%15.10f', save_files=False):
        """
        Get ESP at each point on a grid from a given file or compute it with Psi4.

        Parameters
        ----------
        basis: str (optional)
            Basis set to compute ESP
        method: str (optional)
            Method to compute ESP
        solvent: str (optional)
            Solvent for computing in implicit solvent
        psi4_options: dict (optional)
            additional Psi4 options
        fmt: str
            float format
        save_files: bool (optional)
            if ``True``, Psi4 files are saved and the computed ESP 
            is written to a file.

        Returns
        -------
        grid_esp: ndarray
        """
        if not save_files:
            cwd = os.getcwd()
            with tempfile.TemporaryDirectory(prefix='tmp') as tmpdir:
                os.chdir(tmpdir)
                self._get_esp(method=method, basis=basis,
                              psi4_options=psi4_options, solvent=solvent,
                              fmt=fmt)
                os.chdir(cwd)
        else:
            self._get_esp(method=method, basis=basis,
                          psi4_options=psi4_options, solvent=solvent,
                          fmt=fmt)
        return self.esp

    def _get_esp(self, basis='6-31g*', method='scf', solvent=None,
                 psi4_options={}, fmt='%15.10f'):
        """
        Get ESP at each point on a grid from a given file or compute it with Psi4.

        Parameters
        ----------
        basis: str (optional)
            Basis set to compute ESP
        method: str (optional)
            Method to compute ESP
        solvent: str (optional)
            Solvent for computing in implicit solvent
        psi4_options: dict (optional)
            additional Psi4 options
        fmt: str
            float format

        Returns
        -------
        grid_esp: ndarray
        """
        import psi4
        grid_fn, esp_fn = 'grid.dat', 'grid_esp.dat'
        logfile = self.esp_filename.rsplit('.', maxsplit=1)[0]+'.log'
        if self.bohr:  # I THINK psi4 converts units based on molecule units
            np.savetxt(grid_fn, self.grid/BOHR_TO_ANGSTROM, fmt=fmt)
        else:
            np.savetxt(grid_fn, self.grid, fmt=fmt)

        psi4.set_output_file(logfile)
        psi4.set_options({'basis': basis})
        psi4.set_options(psi4_options)

        msg = 'Computing grid ESP for {} with {}/{}, solvent={}'
        log.debug(msg.format(self.name, method, basis, solvent))

        if solvent:
            psi4.set_options({'pcm': True,
                              'pcm_scf_type': 'total'})
            fname = psi4.core.get_local_option('PCM', 'PCMSOLVER_PARSED_FNAME')
            if not fname or not os.path.exists(fname):
                block = textwrap.dedent("""
                    units = angstrom
                    medium {{
                        solvertype = CPCM
                        solvent = {}
                    }}
                    cavity {{
                        radiiset = bondi  # Bondi | UFF | Allinger
                        type = gepol
                        scaling = True  # radii for spheres scaled by 1.2
                        area = 0.3
                        mode = implicit
                    }}
                """.format(solvent))
                psi4.pcm_helper(block)

        E, wfn = psi4.prop(method, properties=['GRID_ESP'], molecule=self.molecule,
                           return_wfn=True)

        if solvent:
            import pcmsolver  # clear pcmsolver or it's sad the next time
            pcmsolver.getkw.GetkwParser.bnf = None

        self.esp = np.array(wfn.oeprop.Vvals())
        psi4.core.clean()
        psi4.core.clean_options()
        np.savetxt(self.esp_filename, self.esp, fmt=fmt)
        return self.esp



"""
Unit and regression test for the psiresp package.
"""

# Import package, test suite, and other packages as needed
import psiresp
import pytest
import sys
import numpy as np

from numpy.testing import assert_almost_equal, assert_allclose
from .utils import mol_from_file, charges_from_red_file, datafile


class TestNoOrient:

    @pytest.fixture()
    def resp(self):
        confs = [mol_from_file('dmso_opt_c1.xyz')]
        r = psiresp.Resp.from_molecules(confs, charge=0)
        return r

    def test_esp_1(self, resp):
        charges = resp.fit(restraint=False, save_files=False)
        ref = np.array([-0.43877469,  0.14814998,  0.17996033,  0.18716814,  0.35743529,
                        -0.5085439, -0.46067469,  0.19091725,  0.15500465,  0.18935764])
        assert_almost_equal(charges, ref, decimal=5)

    def test_resp_1(self, resp):
        charges = resp.fit(restraint=True)
        ref = np.array([-0.31436216,  0.11376836,  0.14389443,  0.15583112,  0.30951582,
                        -0.50568553, -0.33670393,  0.15982115,  0.12029174,  0.153629])
        assert_almost_equal(charges, ref, decimal=5)

    def test_esp_2(self, resp):
        charges = resp.run(restraint=False, stage_2=True, opt=False)
        ref = np.array([-0.39199538,  0.15716631,  0.15716631,  0.15716631,  0.35743529,
                        -0.5085439, -0.43701446,  0.16953984,  0.16953984,  0.16953984])
        assert_almost_equal(charges, ref, decimal=5)

    def test_resp_2(self, resp):
        charges = resp.run(restraint=True, stage_2=True, opt=False)
        ref = np.array([-0.25158642,  0.11778735,  0.11778735,  0.11778735,  0.30951582,
                        -0.50568553, -0.29298059,  0.12912489,  0.12912489,  0.12912489])
        assert_almost_equal(charges, ref, decimal=5)

    @pytest.mark.parametrize('stage_2,a,redname', [
        (False, 0.01, 'respA2'),
        (True, 0.0005, 'respA1'),
        (False, 0.0, 'espA1')
    ])
    def test_preorient_dmso(self, stage_2, a, redname):
        fns = ['dmso_opt_c1_o1.xyz', 'dmso_opt_c1_o2.xyz']
        mols = [mol_from_file(f) for f in fns]
        r = psiresp.Resp.from_molecules(mols, charge=0)
        charges = r.run(stage_2=stage_2, opt=False, hyp_a1=a, equal_methyls=True)
        ref = charges_from_red_file('dmso_c1_o2_{}.dat'.format(redname))
        assert_allclose(charges, ref, rtol=0.01, atol=1e-4)

    @pytest.mark.parametrize('stage_2,a,redname', [
        (False, 0.01, 'respA2'),
        (True, 0.0005, 'respA1'),
        (False, 0.0, 'espA1')
    ])
    def test_noorient_dmso(self, stage_2, a, redname):
        mols = [mol_from_file('dmso_opt_c1_qmra.xyz')]
        r = psiresp.Resp.from_molecules(mols, charge=0)
        charges = r.run(stage_2=stage_2, opt=False, hyp_a1=a, equal_methyls=True)
        ref = charges_from_red_file('dmso_c1_o0_{}.dat'.format(redname))
        assert_allclose(charges, ref, rtol=0.01, atol=1e-4)


@pytest.mark.parametrize('stage_2,a,redname', [
    (False, 0.01, 'respA2'),
    (True, 0.0005, 'respA1'),
    (False, 0.0, 'espA1')
])
class TestRespNoOpt(object):
    molfile = '{molname}_opt_c{conf}.xyz'
    chargefile = '{molname}_c{n_conf}_o{n_orient}_{name}.dat'
    opt = False
    load_files = False

    GRID = datafile('test_resp/grid.dat')
    ESP = datafile('test_resp/grid_esp.dat')

    def load_mols(self, molname, nconf):
        fns = [self.molfile.format(molname=molname, conf=i+1) for i in range(nconf)]
        return [mol_from_file(f) for f in fns]

    def load_charges(self, molname, n_conf, n_orient, name):
        chargefile = self.chargefile.format(molname=molname, n_conf=n_conf,
                                            n_orient=n_orient, name=name)
        return charges_from_red_file(chargefile)

    def test_resp_single_conf(self, stage_2, a, redname, tmpdir):
        confs = self.load_mols('dmso', 1)
        r = psiresp.Resp.from_molecules(confs, charge=0, name='dmso',
                                        load_files=self.load_files,
                                        grid_name=self.GRID,
                                        esp_name=self.ESP)
        with tmpdir.as_cwd():
            charges = r.run(stage_2=stage_2, opt=self.opt, hyp_a1=a, restraint=True,
                            equal_methyls=True, n_orient=2, save_files=False)
        ref = self.load_charges('dmso', 1, 2, redname,)
        assert_allclose(charges, ref, rtol=0.01, atol=1e-4)

    def test_resp_multi_conf(self, stage_2, a, redname, tmpdir):
        confs = self.load_mols('ethanol', 2)
        r = psiresp.Resp.from_molecules(confs, charge=0, name='ethanol',
                                        orient=[(1, 5, 8), (8, 5, 1),
                                                (9, 8, 5), (5, 8, 9)],
                                        load_files=self.load_files,
                                        grid_name=self.GRID,
                                        esp_name=self.ESP)
        if not stage_2:
            chrequiv = [[2, 3, 4], [6, 7]]
        else:
            chrequiv = []
        with tmpdir.as_cwd():
            charges = r.run(stage_2=stage_2, opt=self.opt, hyp_a1=a,
                            chrequiv=chrequiv, save_files=False)
        ref = self.load_charges('ethanol', 2, 4, redname)
        assert_allclose(charges, ref, rtol=0.01, atol=5e-4)

    @pytest.mark.parametrize('chargename,chrconstr', [
        ('', [(0, [1, 2, 3, 4, 5, 6]),
              (0, [20, 21, 22, 23, 24, 25]),
              (-0.4157, [7]), (0.2719, [8]),
              (0.5973, [18]), (-0.5679, [19]),
              ]),
        # # skip this b/c tests are timing out and it's replicated in MultiResp
        # ('_multifit_constr', {
        #     0: [20, 21, 22, 23, 24, 25],
        #     0.6163: [18],
        #     -0.5722: [19]}),
    ])
    def test_intra_constraints(self, chrconstr, chargename, stage_2, a, redname,
                               tmpdir):
        confs = self.load_mols('nme2ala2', 2)
        chrequiv = [[10, 14], [11, 12, 13, 15, 16, 17]]
        orient = [(5, 18, 19), (19, 18, 5), (6, 19, 20), (20, 19, 6)]
        r = psiresp.Resp.from_molecules(confs, charge=0, name='nme2ala2',
                                        load_files=self.load_files,
                                        grid_name=self.GRID,
                                        esp_name=self.ESP)
        with tmpdir.as_cwd():
            charges = r.run(stage_2=stage_2, opt=self.opt, hyp_a1=a,
                            equal_methyls=False, chrequiv=chrequiv,
                            chrconstr=chrconstr, orient=orient,
                            save_files=False)
        ref = self.load_charges('nme2ala2'+chargename, 2, 4, redname)
        assert_allclose(charges, ref, rtol=0.01, atol=1e-4)


@pytest.mark.fast
class TestLoadNoOpt(TestRespNoOpt):
    load_files = True


@pytest.mark.optimize
@pytest.mark.slow
class TestOpt(TestRespNoOpt):
    molfile = '{molname}_c{conf}.xyz'
    opt = True
